package org.bidib.wizard.comm.bidib;

import java.nio.ByteBuffer;
import java.util.Arrays;
import java.util.BitSet;
import java.util.Calendar;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.bidib.jbidibc.Bidib;
import org.bidib.jbidibc.BidibLibrary;
import org.bidib.jbidibc.Feature;
import org.bidib.jbidibc.FirmwareUpdateStat;
import org.bidib.jbidibc.LcConfig;
import org.bidib.jbidibc.LcMacro;
import org.bidib.jbidibc.Node;
import org.bidib.jbidibc.ProtocolVersion;
import org.bidib.jbidibc.SoftwareVersion;
import org.bidib.jbidibc.VendorData;
import org.bidib.jbidibc.enumeration.AnalogPortEnum;
import org.bidib.jbidibc.enumeration.BidibEnum;
import org.bidib.jbidibc.enumeration.CommandStationState;
import org.bidib.jbidibc.enumeration.CriticalFunctionEnum;
import org.bidib.jbidibc.enumeration.DriveAcknowledge;
import org.bidib.jbidibc.enumeration.FeatureEnum;
import org.bidib.jbidibc.enumeration.FirmwareUpdateOperation;
import org.bidib.jbidibc.enumeration.FlagEnum;
import org.bidib.jbidibc.enumeration.IdentifyState;
import org.bidib.jbidibc.enumeration.InputEnum;
import org.bidib.jbidibc.enumeration.LcMacroOperationCode;
import org.bidib.jbidibc.enumeration.LcMacroState;
import org.bidib.jbidibc.enumeration.LcOutputType;
import org.bidib.jbidibc.enumeration.LightPortEnum;
import org.bidib.jbidibc.enumeration.MacroEnum;
import org.bidib.jbidibc.enumeration.MotorPortEnum;
import org.bidib.jbidibc.enumeration.ServoPortEnum;
import org.bidib.jbidibc.enumeration.SoundPortEnum;
import org.bidib.jbidibc.enumeration.SwitchPortEnum;
import org.bidib.jbidibc.exception.NoAnswerException;
import org.bidib.jbidibc.exception.ProtocolNoAnswerException;
import org.bidib.jbidibc.message.BoostDiagnosticResponse;
import org.bidib.jbidibc.message.SysClockMessage;
import org.bidib.jbidibc.node.AccessoryNode;
import org.bidib.jbidibc.node.BidibNode;
import org.bidib.wizard.comm.AnalogPortStatus;
import org.bidib.wizard.comm.BidibStatus;
import org.bidib.wizard.comm.Communication;
import org.bidib.wizard.comm.CriticalFunctionStatus;
import org.bidib.wizard.comm.Direction;
import org.bidib.wizard.comm.FlagStatus;
import org.bidib.wizard.comm.InputStatus;
import org.bidib.wizard.comm.LightPortStatus;
import org.bidib.wizard.comm.MacroStatus;
import org.bidib.wizard.comm.MotorPortStatus;
import org.bidib.wizard.comm.ServoPortStatus;
import org.bidib.wizard.comm.SoundPortStatus;
import org.bidib.wizard.comm.SpeedSteps;
import org.bidib.wizard.comm.SwitchPortStatus;
import org.bidib.wizard.comm.listener.CommunicationListener;
import org.bidib.wizard.locale.Resources;
import org.bidib.wizard.mvc.main.model.Accessory;
import org.bidib.wizard.mvc.main.model.AnalogPort;
import org.bidib.wizard.mvc.main.model.FeedbackPort;
import org.bidib.wizard.mvc.main.model.Flag;
import org.bidib.wizard.mvc.main.model.InputPort;
import org.bidib.wizard.mvc.main.model.LightPort;
import org.bidib.wizard.mvc.main.model.Macro;
import org.bidib.wizard.mvc.main.model.MacroRepeatDay;
import org.bidib.wizard.mvc.main.model.MacroRepeatTime;
import org.bidib.wizard.mvc.main.model.MotorPort;
import org.bidib.wizard.mvc.main.model.Port;
import org.bidib.wizard.mvc.main.model.ServoPort;
import org.bidib.wizard.mvc.main.model.SoundPort;
import org.bidib.wizard.mvc.main.model.StartCondition;
import org.bidib.wizard.mvc.main.model.SwitchPort;
import org.bidib.wizard.mvc.main.model.TimeStartCondition;
import org.bidib.wizard.mvc.main.model.function.AnalogPortAction;
import org.bidib.wizard.mvc.main.model.function.CriticalFunction;
import org.bidib.wizard.mvc.main.model.function.DelayFunction;
import org.bidib.wizard.mvc.main.model.function.FlagFunction;
import org.bidib.wizard.mvc.main.model.function.Function;
import org.bidib.wizard.mvc.main.model.function.InputFunction;
import org.bidib.wizard.mvc.main.model.function.LightPortAction;
import org.bidib.wizard.mvc.main.model.function.MacroFunction;
import org.bidib.wizard.mvc.main.model.function.MotorPortAction;
import org.bidib.wizard.mvc.main.model.function.PortAction;
import org.bidib.wizard.mvc.main.model.function.RandomDelayFunction;
import org.bidib.wizard.mvc.main.model.function.ServoPortAction;
import org.bidib.wizard.mvc.main.model.function.SoundPortAction;
import org.bidib.wizard.mvc.main.model.function.SwitchPortAction;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class BidibCommunication implements Communication {
    private static final Logger LOGGER = LoggerFactory.getLogger(BidibCommunication.class);

    private final Set<CommunicationListener> communicationListeners = new LinkedHashSet<CommunicationListener>();

    private static final byte[] MACRO_START_OFF = new byte[] { (byte) 0x3F, (byte) 0xBF, (byte) 0x7F, (byte) 0xFF };

    private final Map<Long, Set<Feature>> features = new HashMap<Long, Set<Feature>>();

    private final Map<Long, List<Accessory>> accessories = new HashMap<Long, List<Accessory>>();

    private final Map<Long, List<AnalogPort>> analogPorts = new HashMap<Long, List<AnalogPort>>();

    private final Map<Long, List<FeedbackPort>> feedbackPorts = new HashMap<Long, List<FeedbackPort>>();

    private final List<Flag> flags = new LinkedList<Flag>();

    private final Map<Long, List<InputPort>> inputPorts = new HashMap<Long, List<InputPort>>();

    private final Map<Long, List<LightPort>> lightPorts = new HashMap<Long, List<LightPort>>();

    private final Map<Long, List<Macro>> macros = new HashMap<Long, List<Macro>>();

    private final Map<Long, List<MotorPort>> motorPorts = new HashMap<Long, List<MotorPort>>();

    private final Collection<Node> nodes = new LinkedList<Node>();

    private final Map<Long, List<ServoPort>> servoPorts = new HashMap<Long, List<ServoPort>>();

    private final Map<Long, List<SoundPort>> soundPorts = new HashMap<Long, List<SoundPort>>();

    private final Map<Long, List<SwitchPort>> switchPorts = new HashMap<Long, List<SwitchPort>>();

    private String connectedCommPort;

    /**
     * Creates a new instance of BidibCommunication with the connected comm port.
     */
    public BidibCommunication() {
    }

    @Override
    public void addCommunicationListener(CommunicationListener listener) {
        communicationListeners.add(listener);
    }

    @Override
    public void open(String commPort) {
        LOGGER.info("Try to open the commport: {}", commPort);
        // open the comm port
        try {
            Bidib.getInstance().open(commPort);
        }
        catch (Exception ex) {
            LOGGER.warn("Open port failed.", ex);
            throw new RuntimeException("Open port failed.", ex);
        }

        this.connectedCommPort = commPort;

        fireOpened(commPort);
        // flags
        for (int id = 0; id < 16; id++) {
            Flag flag = new Flag();

            flag.setId(id);
            flags.add(flag);
        }

        LOGGER.info("Get the root node from Bidib. The connected commport is: {}", connectedCommPort);
        reset(null);

        try {
            Thread.sleep(1500);
        }
        catch (InterruptedException e) {
            LOGGER.warn("Wait for bus reset failed.", e);
        }
        LOGGER.info("Reset BiDiBus passed.");
        fireStatus(Resources.getString(getClass(), "bidib-reset-passed"), -1);

        // nodes
        BidibNode rootNode = Bidib.getInstance().getRootNode();
        LOGGER.debug("Fetched rootNode: {}", rootNode);

        try {
            int count = rootNode.getNodeCount();
            LOGGER.debug("Get nodes, count: {}", count);

            for (int index = 1; index <= count; index++) {
                Node node = rootNode.getNextNode();
                LOGGER.debug("Add new node: {}", node);
                nodes.add(node);
                LOGGER.debug("saved nodes: {}", nodes);

                LOGGER.debug("get magic for new node: {}", node);
                Bidib.getInstance().getNode(node).getMagic();
            }
        }
        catch (Exception e) {
            LOGGER.error("Get nodes from Bidib failed.", e);
            throw new RuntimeException(e);
        }
        fireStatus(Resources.getString(getClass(), "bidib-nodes-loaded"), -1);
    }

    @Override
    public void clearCachedData(Node node) {
        LOGGER.debug("Clear all cached data of node: {}", node);

        features.remove(node.getUniqueId());
        accessories.remove(node.getUniqueId());
        analogPorts.remove(node.getUniqueId());
        feedbackPorts.remove(node.getUniqueId());
        inputPorts.remove(node.getUniqueId());
        lightPorts.remove(node.getUniqueId());
        macros.remove(node.getUniqueId());
        motorPorts.remove(node.getUniqueId());
        soundPorts.remove(node.getUniqueId());
        switchPorts.remove(node.getUniqueId());
    }

    public String getConnectedCommPort() {
        return connectedCommPort;
    }

    @Override
    public Set<Feature> getFeatures(Node node) {
        // get features from cache ...
        Set<Feature> result = features.get(node.getUniqueId());
        if (result == null) {
            result = new LinkedHashSet<Feature>();
            features.put(node.getUniqueId(), result);

            try {
                BidibNode bidibNode = Bidib.getInstance().getNode(node);
                int numFeatures = bidibNode.getFeatureCount();
                if (numFeatures > 0) {
                    // get the features
                    Feature feature = null;
                    while ((feature = bidibNode.getNextFeature()) != null) {
                        // add feature to list
                        result.add(feature);
                    }
                }
            }
            catch (Exception ex) {
                LOGGER.warn("Get the features from the noide failed.", ex);
                throw new RuntimeException(ex);
            }
        }
        return result;
    }

    @Override
    public Feature readFeature(Node node, int featureNumber) {
        Feature feature = null;
        // get features from cache ...
        Set<Feature> result = features.get(node.getUniqueId());
        if (result == null) {
            // no cached data found
            result = new LinkedHashSet<Feature>();
            features.put(node.getUniqueId(), result);

            try {
                BidibNode bidibNode = Bidib.getInstance().getNode(node);
                feature = bidibNode.getFeature(featureNumber);
                if (feature != null) {
                    // add feature to list
                    result.add(feature);
                }
            }
            catch (Exception ex) {
                LOGGER.warn("Get the features from the node failed.", ex);
                throw new RuntimeException(ex);
            }
        }
        else {
            // search in the resultset for the cached feature
            for (Feature temp : result) {
                if (temp.getType() == featureNumber) {
                    feature = temp;
                    break;
                }
            }
        }
        return feature;
    }

    @Override
    public void writeFeature(Node node, Feature feature) {

        try {
            BidibNode bidibNode = Bidib.getInstance().getNode(node);
            bidibNode.setFeature(feature.getType(), feature.getValue());

            Set<Feature> result = features.get(node.getUniqueId());
            if (result == null) {
                result = new LinkedHashSet<Feature>();
                features.put(node.getUniqueId(), result);
            }
            // add feature to the cache set
            result.add(feature);
        }
        catch (Exception ex) {
            LOGGER.warn("Write new feature value to the node failed.", ex);
            throw new RuntimeException(ex);
        }

        // some features force internal data to be updated
        // e.g. BiDiB-S88-Brigde allows to configure the number feedback ports
        try {
            switch (FeatureEnum.valueOf((byte) feature.getType())) {
                case FEATURE_BM_SIZE:
                    // clear number of feedback ports
                    LOGGER.debug("remove cached feedback ports for node: {}", node);
                    feedbackPorts.remove(node.getUniqueId());
                    break;
                default:
                    break;
            }
        }
        catch (Exception ex) {
            LOGGER.warn("Update internal data with new feature value failed.", ex);
        }
    }

    @Override
    public void activateAnalogPort(Node node, int port, AnalogPortStatus status) {
        try {
            Bidib.getInstance().getAccessoryNode(node).setOutput(LcOutputType.ANALOGPORT, port,
                status.getType().getType());
        }
        catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public void activateLightPort(Node node, int port, LightPortStatus status) {
        try {
            Bidib.getInstance().getAccessoryNode(node).setOutput(LcOutputType.LIGHTPORT, port,
                status.getType().getType());
        }
        catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public void activateMotorPort(Node node, int port, MotorPortStatus status) {
        try {
            Bidib.getInstance().getAccessoryNode(node).setOutput(LcOutputType.MOTORPORT, port,
                status.getType().getType());
        }
        catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public void activateServoPort(Node node, int port, int value) {
        try {
            Bidib.getInstance().getAccessoryNode(node).setOutput(LcOutputType.SERVOPORT, port, value);
        }
        catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public void activateSoundPort(Node node, int port, SoundPortStatus status) {
        try {
            Bidib.getInstance().getAccessoryNode(node).setOutput(LcOutputType.SOUNDPORT, port,
                status.getType().getType());
        }
        catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public void activateSwitchPort(Node node, int port, SwitchPortStatus status) {
        try {
            Bidib.getInstance().getAccessoryNode(node).setOutput(LcOutputType.SWITCHPORT, port,
                status.getType().getType());
        }
        catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public void boosterOff(Node node) {
        LOGGER.info("Switch only command station off, booster stays active! Selected node: {}", node);
        try {
            Bidib.getInstance().getCommandStationNode(node).setState(CommandStationState.OFF);
        }
        catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public void boosterOn(Node node) {
        LOGGER.info("Switch command station and booster on! Selected node: {}", node);
        try {
            // make sure all locos are removed from loco buffer
            DriveAcknowledge acknowledge = Bidib.getInstance().getCommandStationNode(node).releaseLoco(0);
            LOGGER.info("Clear locobuffer returned: {}", acknowledge);

            // get the current watchdog interval
            Feature watchdogInterval =
                Bidib.getInstance().getNode(node).getFeature(FeatureEnum.FEATURE_GEN_WATCHDOG.getType());
            LOGGER.info("Configured watchdog feature: {}", watchdogInterval);

            // try to start the command station with ignore watchdog
            CommandStationState commandStationState =
                Bidib.getInstance().getCommandStationNode(node).setState(CommandStationState.GO_IGN_WD);
            LOGGER.info("New commandStationState after send GO_IGN_WD: {}", commandStationState);
            if (CommandStationState.OFF.equals(commandStationState)) {
                // no watchdog support available
                commandStationState = Bidib.getInstance().getCommandStationNode(node).setState(CommandStationState.GO);
                LOGGER.info("New commandStationState after send GO: {}", commandStationState);
            }
            Bidib.getInstance().getNode(node).boosterOn();
        }
        catch (Exception e) {
            LOGGER.warn("Switch booster on failed.");
            throw new RuntimeException("Switch booster on failed.", e);
        }
    }

    @Override
    public void boosterQuery(Node node) {
        try {
            Bidib.getInstance().getNode(node).boosterQuery();
        }
        catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public void disable(Node node) {
        try {
            if (node != null) {
                Bidib.getInstance().getNode(node).sysDisable();
            }
            else {
                Bidib.getInstance().getRootNode().sysDisable();
            }
        }
        catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public void enable(Node node) {
        try {
            if (node != null) {
                Bidib.getInstance().getNode(node).sysEnable();
            }
            else {
                Bidib.getInstance().getRootNode().sysEnable();
            }
        }
        catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public Collection<Accessory> getAccessories(Node node) {
        List<Accessory> result = accessories.get(node.getUniqueId());

        if (result == null) {
            result = new LinkedList<Accessory>();
            accessories.put(node.getUniqueId(), result);
            try {
                AccessoryNode accessoryNode = Bidib.getInstance().getAccessoryNode(node);

                if (accessoryNode != null) {
                    Feature accessoryCount = accessoryNode.getFeature(BidibLibrary.FEATURE_ACCESSORY_COUNT);

                    if (accessoryCount != null) {
                        for (int accessoryNumber = 0; accessoryNumber < accessoryCount.getValue(); accessoryNumber++) {
                            Accessory accessory = new Accessory();

                            accessory.setId(accessoryNumber);
                            getAccessoryContent(node, accessory);
                            result.add(accessory);
                        }
                    }
                }
            }
            catch (Exception e) {
                LOGGER.warn("Get accessories failed.", e);
                throw new RuntimeException(e);
            }
        }
        return result;
    }

    private void getAccessoryContent(Node node, Accessory accessory) {
        try {
            AccessoryNode accessoryNode = Bidib.getInstance().getAccessoryNode(node);

            if (accessoryNode != null) {
                Feature macroMapped = accessoryNode.getFeature(BidibLibrary.FEATURE_ACCESSORY_MACROMAPPED);

                if (macroMapped != null && macroMapped.getValue() > 0) {
                    accessory.setMacroSize(macroMapped.getValue());

                    byte[] parameter =
                        accessoryNode.getAccessoryParameter(accessory.getId(),
                            BidibLibrary.BIDIB_ACCESSORY_PARA_MACROMAP);

                    if (parameter != null) {
                        List<Macro> macroList = internalGetMacros(node);

                        accessory.clearMacros();
                        for (int aspectNumber = 0; aspectNumber < parameter.length; aspectNumber++) {
                            int macroNumber = parameter[aspectNumber];

                            if (macroNumber >= 0 && macroNumber < macroList.size() && macroNumber != 255) {
                                accessory.addMacro(macroList.get(macroNumber));
                            }
                        }
                    }
                }
            }
        }
        catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public int getAccessoryLength(Node node) {
        int result = 0;

        try {
            AccessoryNode accessoryNode = Bidib.getInstance().getAccessoryNode(node);

            if (accessoryNode != null) {
                Feature feature = accessoryNode.getFeature(BidibLibrary.FEATURE_ACCESSORY_MACROMAPPED);

                result = feature != null ? feature.getValue() : 0;
            }
        }
        catch (Exception e) {
            throw new RuntimeException(e);
        }
        return result;
    }

    @Override
    public List<AnalogPort> getAnalogPorts(Node node) {
        List<AnalogPort> result = analogPorts.get(node.getUniqueId());

        if (result == null) {
            result = new LinkedList<AnalogPort>();
            analogPorts.put(node.getUniqueId(), result);
            try {
                AccessoryNode accessoryNode = Bidib.getInstance().getAccessoryNode(node);

                if (accessoryNode != null) {
                    Feature feature = accessoryNode.getFeature(BidibLibrary.FEATURE_CTRL_ANALOG_COUNT);

                    if (feature != null) {
                        for (int index = 0; index < feature.getValue(); index++) {
                            AnalogPort port = new AnalogPort();

                            port.setId(index);
                            result.add(port);
                        }
                    }
                }
            }
            catch (Exception e) {
                throw new RuntimeException(e);
            }
        }
        return Collections.unmodifiableList(result);
    }

    @Override
    public int getBoosterMaximumCurrent(Node node) {
        int result = 0;

        try {
            Feature feature = Bidib.getInstance().getNode(node).getFeature(BidibLibrary.FEATURE_BST_AMPERE);

            if (feature != null) {
                result = BoostDiagnosticResponse.convertCurrent(feature.getValue());
            }
        }
        catch (Exception e) {
            throw new RuntimeException(e);
        }
        return result;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public Collection<FeedbackPort> getFeedbackPorts(Node node) {
        List<FeedbackPort> result = feedbackPorts.get(node.getUniqueId());

        if (result == null) {
            result = new LinkedList<FeedbackPort>();
            feedbackPorts.put(node.getUniqueId(), result);
            try {
                BidibNode bidibNode = Bidib.getInstance().getNode(node);

                if (bidibNode != null) {
                    // get the number of feedback ports from the system
                    Feature feature = bidibNode.getFeature(BidibLibrary.FEATURE_BM_SIZE);

                    if (feature != null) {
                        LOGGER.debug("Number of feedback ports: {}", feature.getValue());
                        for (int index = 0; index < feature.getValue(); index++) {
                            FeedbackPort port = new FeedbackPort();

                            port.setId(index);
                            result.add(port);
                        }
                    }
                }
                else {
                    LOGGER.warn("Requested node is not available in system.");
                }
            }
            catch (Exception e) {
                LOGGER.warn("Get requested node from system failed: {}", node, e);
                throw new RuntimeException(e);
            }
        }
        return Collections.unmodifiableList(result);
    }

    @Override
    public void getFeedbackPortStatus(Node node, int begin, int end) {
        try {
            BidibNode bidibNode = Bidib.getInstance().getNode(node);

            bidibNode.getFeedbackState(begin, end);
            bidibNode.getAddressState(begin, end);
            bidibNode.getConfidence();
        }
        catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public List<Flag> getFlags() {
        return Collections.unmodifiableList(flags);
    }

    private Function getFunction(LcOutputType outputType, BidibStatus status) {
        Function result = null;

        if (outputType == LcOutputType.ANALOGPORT) {
            result = new AnalogPortAction();
            result.setAction(status);
        }
        else if (outputType == LcOutputType.LIGHTPORT) {
            result = new LightPortAction();
            result.setAction(status);
        }
        else if (outputType == LcOutputType.MOTORPORT) {
            result = new MotorPortAction();
            result.setAction(status);
        }
        else if (outputType == LcOutputType.SERVOPORT) {
            result = new ServoPortAction();
            result.setAction(status);
        }
        else if (outputType == LcOutputType.SOUNDPORT) {
            result = new SoundPortAction();
            result.setAction(status);
        }
        else if (outputType == LcOutputType.SWITCHPORT) {
            result = new SwitchPortAction();
            result.setAction(status);
        }
        else if (outputType == LcOutputType.DELAY) {
            result = new DelayFunction();
        }
        else if (outputType == LcOutputType.RANDOM_DELAY) {
            result = new RandomDelayFunction();
        }
        else if (outputType == LcOutputType.INPUT_QUERY0) {
            result = new InputFunction();
            result.setAction(InputStatus.QUERY0);
        }
        else if (outputType == LcOutputType.INPUT_QUERY1) {
            result = new InputFunction();
            result.setAction(InputStatus.QUERY1);
        }
        else if (outputType == LcOutputType.FLAG_CLEAR) {
            result = new FlagFunction();
            result.setAction(FlagStatus.CLEAR);
        }
        else if (outputType == LcOutputType.FLAG_SET) {
            result = new FlagFunction();
            result.setAction(FlagStatus.SET);
        }
        else if (outputType == LcOutputType.FLAG_QUERY) {
            result = new FlagFunction();
            result.setAction(FlagStatus.QUERY);
        }
        else if (outputType == LcOutputType.END_CRITICAL) {
            result = new CriticalFunction();
            result.setAction(CriticalFunctionStatus.END);
        }
        else if (outputType == LcOutputType.BEGIN_CRITICAL) {
            result = new CriticalFunction();
            result.setAction(CriticalFunctionStatus.BEGIN);
        }
        else if (outputType == LcOutputType.STOP_MACRO) {
            result = new MacroFunction();
            result.setAction(MacroStatus.STOP);
        }
        else if (outputType == LcOutputType.START_MACRO) {
            result = new MacroFunction();
            result.setAction(MacroStatus.START);
        }
        else if (outputType == LcOutputType.END_OF_MACRO) {
            result = new MacroFunction();
            result.setAction(MacroStatus.END);
        }
        else {
            throw new RuntimeException("cannot map output type " + outputType.getClass().getSimpleName()
                + " to a function");
        }
        return result;
    }

    @Override
    public List<InputPort> getInputPorts(Node node) {
        List<InputPort> result = inputPorts.get(node.getUniqueId());

        if (result == null) {
            result = new LinkedList<InputPort>();
            inputPorts.put(node.getUniqueId(), result);
            try {
                AccessoryNode accessoryNode = Bidib.getInstance().getAccessoryNode(node);

                if (accessoryNode != null) {
                    Feature feature = accessoryNode.getFeature(BidibLibrary.FEATURE_CTRL_INPUT_COUNT);

                    if (feature != null) {
                        for (int keyNumber = 0; keyNumber < feature.getValue(); keyNumber++) {
                            InputPort port = new InputPort();

                            port.setId(keyNumber);
                            result.add(port);
                        }
                    }
                }
            }
            catch (Exception e) {
                throw new RuntimeException(e);
            }
        }
        return Collections.unmodifiableList(result);
    }

    @Override
    public void getInputPortStatus(Node node, int port) {
        try {
            Bidib.getInstance().getAccessoryNode(node).getKeyState(port);
        }
        catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public List<LightPort> getLightPorts(Node node) {
        LOGGER.debug("Get the light ports from node: {}", node);
        List<LightPort> result = lightPorts.get(node.getUniqueId());
        LOGGER.debug("Fetched the cached light ports: {}", result);

        if (result == null) {
            LOGGER.debug("Fetch lightports from Bidib for node: {}", node);
            result = new LinkedList<LightPort>();
            lightPorts.put(node.getUniqueId(), result);
            try {
                AccessoryNode accessoryNode = Bidib.getInstance().getAccessoryNode(node);

                if (accessoryNode != null) {
                    Feature feature = accessoryNode.getFeature(BidibLibrary.FEATURE_CTRL_LPORT_COUNT);

                    if (feature != null) {
                        for (int index = 0; index < feature.getValue(); index++) {
                            LightPort port = new LightPort();

                            port.setId(index);

                            LcConfig config = accessoryNode.getConfig(LcOutputType.LIGHTPORT, index);

                            if (config != null) {
                                port.setDimMax(config.getValue3());
                                port.setDimMin(config.getValue4());
                                port.setPwmMax(config.getValue2());
                                port.setPwmMin(config.getValue1());
                            }
                            result.add(port);
                        }
                    }
                }
                else {
                    LOGGER.debug("The node '{}' is not an accessory node.", node);
                }
            }
            catch (Exception e) {
                throw new RuntimeException(e);
            }
        }
        return Collections.unmodifiableList(result);
    }

    @Override
    public Macro getMacroContent(Node node, Macro macro) {
        if (macro.getFunctionCount() == 0) {
            try {
                AccessoryNode accessoryNode = Bidib.getInstance().getAccessoryNode(node);

                if (accessoryNode != null) {
                    macro.clearFunctions();

                    // ensure servo port parameters are available to calculate
                    // the relative servo position
                    getServoPorts(node);

                    Collection<Function> functions = new LinkedList<Function>();
                    int stepNumber = 0;

                    for (;;) {
                        final LcMacro macroStep = accessoryNode.getMacroStep(macro.getId(), stepNumber++);

                        if (macroStep.getOutputType() == LcOutputType.END_OF_MACRO
                            || stepNumber > macro.getFunctionSize()) {
                            break;
                        }

                        Function function = getFunction(macroStep.getOutputType(), getStatus(macroStep.getStatus()));

                        if (function instanceof DelayFunction) {
                            ((DelayFunction) function).setDelay(macroStep.getOutputNumber() & 0xFF);
                        }
                        if (function instanceof FlagFunction) {
                            List<Flag> flags = getFlags();
                            int id = macroStep.getOutputNumber();

                            if (id >= 0 && id < flags.size()) {
                                ((FlagFunction) function).setFlag(flags.get(id));
                            }
                        }
                        else if (function instanceof InputFunction) {
                            List<InputPort> inputPorts = getInputPorts(node);
                            int id = macroStep.getOutputNumber();

                            if (id >= 0 && id < inputPorts.size()) {
                                ((InputFunction) function).setInput(inputPorts.get(id));
                            }
                        }
                        else if (function instanceof MacroFunction) {
                            ((MacroFunction) function).setMacroId(macroStep.getOutputNumber());
                        }
                        else if (function instanceof PortAction) {
                            ((PortAction) function).setDelay(macroStep.getDelay() & 0xFF);
                            ((PortAction) function).setPort(getPort(node, macroStep.getOutputType(), macroStep
                                .getOutputNumber()));
                            if (function instanceof ServoPortAction) {
                                ((ServoPortAction) function).setValue(macroStep.getValue() & 0xFF);
                            }
                        }
                        else if (function instanceof RandomDelayFunction) {
                            ((RandomDelayFunction) function).setMaximumValue(macroStep.getOutputNumber() & 0xFF);
                        }
                        functions.add(function);
                    }
                    macro.addFunctionsBefore(0, functions.toArray(new Function[0]));

                    // read the start condition
                    byte[] startClk =
                        accessoryNode.getMacroParameter(macro.getId(), BidibLibrary.BIDIB_MACRO_PARA_START_CLK);
                    applyMacroStartClk(startClk, macro);
                }
            }
            catch (Exception e) {
                throw new RuntimeException(e);
            }
        }
        return macro;
    }

    @Override
    public int getMacroLength(Node node) {
        int result = 0;

        try {
            AccessoryNode accessoryNode = Bidib.getInstance().getAccessoryNode(node);

            if (accessoryNode != null) {
                // http://www.opendcc.de/elektronik/lightcontrol1/lightcontrol.html
                Feature feature = accessoryNode.getFeature(BidibLibrary.FEATURE_CTRL_MAC_SIZE);
                LOGGER.debug("returned macro length feature: {}", feature);
                result = feature != null ? feature.getValue() : 0;
            }
            else {
                LOGGER.debug("The current node is not an AccessoryNode: {}", node);
            }
        }
        catch (Exception e) {
            throw new RuntimeException(e);
        }
        return result;
    }

    @Override
    public Collection<Macro> getMacros(Node node) {
        return Collections.unmodifiableCollection(internalGetMacros(node));
    }

    @Override
    public List<MotorPort> getMotorPorts(Node node) {
        List<MotorPort> result = motorPorts.get(node.getUniqueId());

        if (result == null) {
            result = new LinkedList<MotorPort>();
            motorPorts.put(node.getUniqueId(), result);
            try {
                AccessoryNode accessoryNode = Bidib.getInstance().getAccessoryNode(node);

                if (accessoryNode != null) {
                    Feature feature = accessoryNode.getFeature(BidibLibrary.FEATURE_CTRL_MOTOR_COUNT);

                    if (feature != null) {
                        for (int index = 0; index < feature.getValue(); index++) {
                            MotorPort port = new MotorPort();

                            port.setId(index);
                            result.add(port);
                        }
                    }
                }
            }
            catch (Exception e) {
                throw new RuntimeException(e);
            }
        }
        return Collections.unmodifiableList(result);
    }

    @Override
    public Collection<Node> getNodes() {
        return Collections.unmodifiableCollection(nodes);
    }

    private LcOutputType getOutputType(Function function) {
        LcOutputType result = null;

        if (function instanceof AnalogPortAction) {
            result = LcOutputType.ANALOGPORT;
        }
        else if (function instanceof CriticalFunction && function.getAction() == CriticalFunctionStatus.END) {
            result = LcOutputType.END_CRITICAL;
        }
        else if (function instanceof CriticalFunction && function.getAction() == CriticalFunctionStatus.BEGIN) {
            result = LcOutputType.BEGIN_CRITICAL;
        }
        else if (function instanceof DelayFunction) {
            result = LcOutputType.DELAY;
        }
        else if (function instanceof FlagFunction && function.getAction() == FlagStatus.CLEAR) {
            result = LcOutputType.FLAG_CLEAR;
        }
        else if (function instanceof FlagFunction && function.getAction() == FlagStatus.SET) {
            result = LcOutputType.FLAG_SET;
        }
        else if (function instanceof FlagFunction && function.getAction() == FlagStatus.QUERY) {
            result = LcOutputType.FLAG_QUERY;
        }
        else if (function instanceof InputFunction && function.getAction() == InputStatus.QUERY0) {
            result = LcOutputType.INPUT_QUERY0;
        }
        else if (function instanceof InputFunction && function.getAction() == InputStatus.QUERY1) {
            result = LcOutputType.INPUT_QUERY1;
        }
        else if (function instanceof LightPortAction) {
            result = LcOutputType.LIGHTPORT;
        }
        else if (function instanceof MacroFunction && function.getAction() == MacroStatus.STOP) {
            result = LcOutputType.STOP_MACRO;
        }
        else if (function instanceof MacroFunction && function.getAction() == MacroStatus.START) {
            result = LcOutputType.START_MACRO;
        }
        else if (function instanceof MacroFunction && function.getAction() == MacroStatus.END) {
            result = LcOutputType.END_OF_MACRO;
        }
        else if (function instanceof MotorPortAction) {
            result = LcOutputType.MOTORPORT;
        }
        else if (function instanceof RandomDelayFunction) {
            result = LcOutputType.RANDOM_DELAY;
        }
        else if (function instanceof ServoPortAction) {
            result = LcOutputType.SERVOPORT;
        }
        else if (function instanceof SoundPortAction) {
            result = LcOutputType.SOUNDPORT;
        }
        else if (function instanceof SwitchPortAction) {
            result = LcOutputType.SWITCHPORT;
        }
        else {
            throw new RuntimeException("cannot map function type " + function.getClass().getSimpleName()
                + " to an output type");
        }
        return result;
    }

    private Port getPort(Node node, LcOutputType outputType, int outputNumber) {
        Port result = null;
        List<? extends Port> ports = null;

        if (outputType == LcOutputType.ANALOGPORT) {
            ports = getAnalogPorts(node);
        }
        else if (outputType == LcOutputType.LIGHTPORT) {
            ports = getLightPorts(node);
        }
        else if (outputType == LcOutputType.MOTORPORT) {
            ports = getMotorPorts(node);
        }
        else if (outputType == LcOutputType.SERVOPORT) {
            ports = getServoPorts(node);
        }
        else if (outputType == LcOutputType.SOUNDPORT) {
            ports = getSoundPorts(node);
        }
        else if (outputType == LcOutputType.SWITCHPORT) {
            ports = getSwitchPorts(node);
        }
        else {
            throw new RuntimeException("cannot map output type " + outputType.getClass().getSimpleName() + " to a port");
        }
        if (ports != null && outputNumber >= 0 && outputNumber < ports.size()) {
            result = ports.get(outputNumber);
            result.setId(outputNumber);
        }
        return result;
    }

    @Override
    public ProtocolVersion getProtocolVersion(Node node) {
        ProtocolVersion result = null;

        try {
            result = Bidib.getInstance().getNode(node).getPVersion();
        }
        catch (Exception e) {
            throw new RuntimeException(e);
        }
        return result;
    }

    @Override
    public List<ServoPort> getServoPorts(Node node) {
        List<ServoPort> result = servoPorts.get(node.getUniqueId());

        if (result == null) {
            result = new LinkedList<ServoPort>();
            servoPorts.put(node.getUniqueId(), result);
            try {
                AccessoryNode accessoryNode = Bidib.getInstance().getAccessoryNode(node);

                if (accessoryNode != null) {
                    Feature feature = accessoryNode.getFeature(BidibLibrary.FEATURE_CTRL_SERVO_COUNT);

                    if (feature != null) {
                        for (int index = 0; index < feature.getValue(); index++) {
                            ServoPort port = new ServoPort();

                            port.setId(index);

                            LcConfig config = accessoryNode.getConfig(LcOutputType.SERVOPORT, index);

                            if (config != null) {
                                port.setSpeed(config.getValue3());
                                port.setTrimDown(config.getValue1());
                                port.setTrimUp(config.getValue2());
                            }
                            result.add(port);
                        }
                    }
                }
            }
            catch (Exception e) {
                throw new RuntimeException(e);
            }
        }
        return Collections.unmodifiableList(result);
    }

    @Override
    public SoftwareVersion getSoftwareVersion(Node node) {
        SoftwareVersion result = null;

        try {
            result = Bidib.getInstance().getNode(node).getSwVersion();
        }
        catch (Exception e) {
            throw new RuntimeException(e);
        }
        return result;
    }

    @Override
    public List<SoundPort> getSoundPorts(Node node) {
        List<SoundPort> result = soundPorts.get(node.getUniqueId());

        if (result == null) {
            result = new LinkedList<SoundPort>();
            soundPorts.put(node.getUniqueId(), result);
            try {
                AccessoryNode accessoryNode = Bidib.getInstance().getAccessoryNode(node);

                if (accessoryNode != null) {
                    Feature feature = accessoryNode.getFeature(BidibLibrary.FEATURE_CTRL_SOUND_COUNT);

                    if (feature != null) {
                        for (int index = 0; index < feature.getValue(); index++) {
                            SoundPort port = new SoundPort();

                            port.setId(index);
                            result.add(port);
                        }
                    }
                }
            }
            catch (Exception e) {
                throw new RuntimeException(e);
            }
        }
        return Collections.unmodifiableList(result);
    }

    /**
     * Map status between BiDiB type and LightControl type, which also contains a localized label.
     * 
     * @param status
     *            status from BiDiB library
     * 
     * @return LightControl status object
     */
    private BidibStatus getStatus(BidibEnum status) {
        BidibStatus result = null;

        if (status != null) {
            if (status instanceof AnalogPortEnum) {
                result = AnalogPortStatus.valueOf((AnalogPortEnum) status);
            }
            else if (status instanceof CriticalFunctionEnum) {
                result = CriticalFunctionStatus.valueOf((CriticalFunctionEnum) status);
            }
            else if (status instanceof FlagEnum) {
                result = FlagStatus.valueOf((FlagEnum) status);
            }
            else if (status instanceof InputEnum) {
                result = InputStatus.valueOf((InputEnum) status);
            }
            else if (status instanceof LightPortEnum) {
                result = LightPortStatus.valueOf((LightPortEnum) status);
            }
            else if (status instanceof MacroEnum) {
                result = MacroStatus.valueOf((MacroEnum) status);
            }
            else if (status instanceof MotorPortEnum) {
                result = MotorPortStatus.valueOf((MotorPortEnum) status);
            }
            else if (status instanceof ServoPortEnum) {
                result = ServoPortStatus.valueOf((ServoPortEnum) status);
            }
            else if (status instanceof SoundPortEnum) {
                result = SoundPortStatus.valueOf((SoundPortEnum) status);
            }
            else if (status instanceof SwitchPortEnum) {
                result = SwitchPortStatus.valueOf((SwitchPortEnum) status);
            }
            else {
                throw new RuntimeException("cannot map " + status.getClass().getSimpleName() + " to a status");
            }
        }
        return result;
    }

    @Override
    public int getStorableMacroCount(Node commNode) {
        int result = 0;

        try {
            AccessoryNode node = Bidib.getInstance().getAccessoryNode(commNode);

            if (node != null) {
                Feature feature = node.getFeature(BidibLibrary.FEATURE_CTRL_MAC_SAVE);

                result = feature != null ? feature.getValue() : 0;
            }
        }
        catch (Exception e) {
            throw new RuntimeException(e);
        }
        return result;
    }

    @Override
    public List<SwitchPort> getSwitchPorts(Node node) {
        List<SwitchPort> result = switchPorts.get(node.getUniqueId());

        if (result == null) {
            result = new LinkedList<SwitchPort>();
            switchPorts.put(node.getUniqueId(), result);
            try {
                AccessoryNode accessoryNode = Bidib.getInstance().getAccessoryNode(node);

                if (accessoryNode != null) {
                    Feature feature = accessoryNode.getFeature(BidibLibrary.FEATURE_CTRL_SPORT_COUNT);

                    if (feature != null) {
                        for (int index = 0; index < feature.getValue(); index++) {
                            SwitchPort port = new SwitchPort();

                            port.setId(index);
                            result.add(port);
                        }
                    }
                }
            }
            catch (Exception e) {
                throw new RuntimeException(e);
            }
        }
        return Collections.unmodifiableList(result);
    }

    private byte[] getUniqueId(long uniqueId) {
        byte[] result = new byte[7];
        ByteBuffer bb = ByteBuffer.allocate(8);

        bb.putLong(uniqueId);
        System.arraycopy(bb.array(), 1, result, 0, result.length);
        return result;
    }

    @Override
    public void identify(Node node, IdentifyState state) {
        try {
            Bidib.getInstance().getNode(node).identify(state);
        }
        catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    private List<Macro> internalGetMacros(Node node) {
        List<Macro> result = macros.get(node.getUniqueId());

        if (result == null) {
            result = new LinkedList<Macro>();
            macros.put(node.getUniqueId(), result);
            try {
                AccessoryNode accessoryNode = Bidib.getInstance().getAccessoryNode(node);

                if (accessoryNode != null) {
                    Feature feature = accessoryNode.getFeature(BidibLibrary.FEATURE_CTRL_MAC_COUNT);
                    int macroCount = feature != null ? feature.getValue() : 0;
                    int macroLength = getMacroLength(node);

                    for (int macroNumber = 0; macroNumber < macroCount; macroNumber++) {
                        Macro macro = new Macro(macroLength);

                        byte[] startClk =
                            accessoryNode.getMacroParameter(macroNumber, BidibLibrary.BIDIB_MACRO_PARA_START_CLK);
                        applyMacroStartClk(startClk, macro);
                        macro.setCycles(accessoryNode.getMacroParameter(macroNumber,
                            BidibLibrary.BIDIB_MACRO_PARA_REPEAT)[0]);
                        macro.setId(macroNumber);
                        macro.setSpeed(accessoryNode.getMacroParameter(macroNumber,
                            BidibLibrary.BIDIB_MACRO_PARA_SLOWDOWN)[0]);
                        result.add(macro);
                    }
                }
            }
            catch (Exception e) {
                throw new RuntimeException(e);
            }
        }
        return result;
    }

    /**
     * Apply the start clk data from the device to the macro
     * 
     * @param startClk
     *            the start clk data
     * @param macro
     *            the macro instance
     */
    private void applyMacroStartClk(byte[] startClk, Macro macro) {

        if (startClk != null && !Arrays.equals(MACRO_START_OFF, startClk)) {
            int hour = startClk[1] & 0x1F;
            int minute = startClk[0] & 0x3F;
            int day = startClk[2] & 0x7;
            Calendar time = Calendar.getInstance();
            TimeStartCondition startCondition = new TimeStartCondition();

            if (hour == 24) {
                startCondition.setRepeatTime(MacroRepeatTime.HOURLY);
                time.set(Calendar.HOUR_OF_DAY, 0);
            }
            else {
                time.set(Calendar.HOUR_OF_DAY, hour);
            }
            if (minute == 60) {
                startCondition.setRepeatTime(MacroRepeatTime.MINUTELY);
                time.set(Calendar.MINUTE, 0);
            }
            else if (minute == 61) {
                startCondition.setRepeatTime(MacroRepeatTime.HALF_HOURLY);
                time.set(Calendar.MINUTE, 0);
            }
            else if (minute == 62) {
                startCondition.setRepeatTime(MacroRepeatTime.QUARTER_HOURLY);
                time.set(Calendar.MINUTE, 0);
            }
            else {
                time.set(Calendar.MINUTE, minute);
            }
            // day handling
            if (day == 7) {
                // every day
                startCondition.setRepeatDay(MacroRepeatDay.ALL);
            }
            else {
                int d = (day + 2) % 7;
                if (d == 0) {
                    d = 7;
                }
                startCondition.setRepeatDay(MacroRepeatDay.values()[d]);
            }
            time.set(Calendar.SECOND, 0);
            startCondition.setTime(time);
            macro.addStartCondition(startCondition);
        }
    }

    @Override
    public Boolean isAddressMessagesEnabled(Node commNode) {
        Boolean result = null;

        try {
            BidibNode node = Bidib.getInstance().getNode(commNode);

            if (node != null) {
                Feature feature = node.getFeature(BidibLibrary.FEATURE_BM_ADDR_DETECT_ON);

                result = feature != null && feature.getValue() > 0;
            }
        }
        catch (Exception e) {
            throw new RuntimeException(e);
        }
        return result;
    }

    @Override
    public Boolean isCommandStation(Node commNode) {
        LOGGER.debug("Check if node is command station node: {}", commNode);

        // BidibNode bidibNode = Bidib.getNode(commNode);
        // boolean isCSNode = bidibNode instanceof CommandStationNode;
        boolean isCSNode = commNode.hasCommandStationFunctions();

        LOGGER.debug("Node is command station node: {}", isCSNode);
        return isCSNode;
    }

    @Override
    public Boolean isDccStartEnabled(Node node) {
        Boolean result = null;

        try {
            AccessoryNode accessoryNode = Bidib.getInstance().getAccessoryNode(node);

            if (accessoryNode != null) {
                Feature feature = accessoryNode.getFeature(BidibLibrary.FEATURE_CTRL_MAC_START_DCC);

                result = feature != null && feature.getValue() > 0;
            }
        }
        catch (Exception e) {
            throw new RuntimeException(e);
        }
        return result;
    }

    @Override
    public Boolean isExternalStartEnabled(Node node) {
        Boolean result = null;

        try {
            AccessoryNode accessoryNode = Bidib.getInstance().getAccessoryNode(node);

            if (accessoryNode != null) {
                Feature feature = accessoryNode.getFeature(BidibLibrary.FEATURE_CTRL_MAC_START_MAN);

                result = feature != null && feature.getValue() > 0;
            }
        }
        catch (Exception e) {
            throw new RuntimeException(e);
        }
        return result;
    }

    @Override
    public Boolean isFeedbackMessagesEnabled(Node commNode) {
        Boolean result = null;

        try {
            BidibNode node = Bidib.getInstance().getNode(commNode);

            if (node != null) {
                Feature feature = node.getFeature(BidibLibrary.FEATURE_BM_ON);

                result = feature != null && feature.getValue() > 0;
            }
        }
        catch (Exception e) {
            throw new RuntimeException(e);
        }
        return result;
    }

    @Override
    public Boolean isKeyMessagesEnabled(Node node) {
        Boolean result = null;

        try {
            AccessoryNode accessoryNode = Bidib.getInstance().getAccessoryNode(node);

            if (accessoryNode != null) {
                Feature feature = accessoryNode.getFeature(BidibLibrary.FEATURE_CTRL_INPUT_NOTIFY);

                result = feature != null && feature.getValue() > 0;
            }
        }
        catch (Exception e) {
            throw new RuntimeException(e);
        }
        return result;
    }

    @Override
    public boolean isUpdatable(Node node) {
        boolean result = false;

        try {
            result = Bidib.getInstance().getNode(node).isUpdatable();
        }
        catch (Exception e) {
            throw new RuntimeException(e);
        }
        return result;
    }

    @Override
    public int readCv(Node node, int cvNumber) {
        VendorData result = null;

        boolean vendorEnabled = false;
        try {
            vendorEnabled = Bidib.getInstance().getNode(node).vendorEnable(getUniqueId(node.getUniqueId()));
            if (vendorEnabled) {
                result = Bidib.getInstance().getNode(node).vendorGet(String.valueOf(cvNumber));
            }
        }
        catch (Exception e) {
            throw new RuntimeException(e);
        }
        finally {
            if (vendorEnabled) {
                try {
                    Bidib.getInstance().getNode(node).vendorDisable();
                }
                catch (Exception e) {
                    throw new RuntimeException(e);
                }
            }
        }

        try {
            return Integer.parseInt(result.getValue());
        }
        catch (NumberFormatException e) {
            return 0;
        }
    }

    @Override
    public void reloadAccessory(Node node, Accessory accessory) {
        getAccessoryContent(node, accessory);
    }

    @Override
    public void reloadMacro(Node node, Macro macro) {
        try {
            Bidib.getInstance().getAccessoryNode(node).handleMacro(macro.getId(), LcMacroOperationCode.RESTORE);
            macro.clearFunctions();
            getMacroContent(node, macro);
        }
        catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public void reset(Node node) {
        fireStatus(Resources.getString(getClass(), "bidib-reset"), -1);

        try {
            if (node == null) {
                Bidib.getInstance().getRootNode().reset();
            }
            else {
                Bidib.getInstance().getNode(node).reset();
            }
        }
        catch (Exception e) {
            LOGGER.warn("Reset node failed.", e);
            throw new RuntimeException(e);
        }
    }

    @Override
    public void saveAccessory(Node node, Accessory accessory) {
        try {
            List<Integer> macros = accessory.getMacros();
            byte[] value = new byte[macros.size() + 1];

            for (int index = 0; index < macros.size(); index++) {
                value[index] = (byte) macros.get(index).intValue();
            }
            value[value.length - 1] = (byte) 255;
            Bidib.getInstance().getAccessoryNode(node).setAccessoryParameter(accessory.getId(),
                BidibLibrary.BIDIB_ACCESSORY_PARA_MACROMAP, value);
        }
        catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public void saveMacro(Node node, Macro macro) {
        try {
            transferMacro(node, macro);
            Bidib.getInstance().getAccessoryNode(node).handleMacro(macro.getId(), LcMacroOperationCode.SAVE);
        }
        catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public FirmwareUpdateStat sendFirmwareUpdateOperation(Node node, FirmwareUpdateOperation operation, byte... data) {
        FirmwareUpdateStat result = null;

        try {
            result = Bidib.getInstance().getNode(node).sendFirmwareUpdateOperation(operation, data);
        }
        catch (ProtocolNoAnswerException e) {
            throw new NoAnswerException("Send firmware update failed for operation: " + operation.name(), e);
        }
        catch (Exception e) {
            throw new RuntimeException("Send firmware update failed for operation: " + operation.name(), e);
        }
        return result;
    }

    @Override
    public void sendTime(Calendar time, int factor) {
        try {
            Bidib.getInstance().getRootNode().clock(time, factor);
        }
        catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public void setAddressMessagesEnabled(Node node, boolean enabled) {
        try {
            Bidib.getInstance().getNode(node).setFeature(BidibLibrary.FEATURE_BM_ADDR_DETECT_ON, (enabled ? 1 : 0));
        }
        catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public void setBinaryState(Node node, int address, int state, boolean value) {
        try {
            Bidib.getInstance().getCommandStationNode(node).setBinaryState(address, state, value);
        }
        catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public void setDccStartEnabled(Node node, boolean enabled) {
        try {
            Bidib.getInstance().getAccessoryNode(node).setFeature(BidibLibrary.FEATURE_CTRL_MAC_START_DCC,
                (enabled ? 1 : 0));
        }
        catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public void setExternalStartEnabled(Node node, boolean enabled) {
        try {
            Bidib.getInstance().getAccessoryNode(node).setFeature(BidibLibrary.FEATURE_CTRL_MAC_START_MAN,
                (enabled ? 1 : 0));
        }
        catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public void setFeedbackMessagesEnabled(Node node, boolean enabled) {
        try {
            Bidib.getInstance().getNode(node).setFeature(BidibLibrary.FEATURE_BM_ON, (enabled ? 1 : 0));
        }
        catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public void setKeyMessagesEnabled(Node node, boolean enabled) {
        try {
            Bidib.getInstance().getAccessoryNode(node).setFeature(BidibLibrary.FEATURE_CTRL_INPUT_NOTIFY,
                (enabled ? 1 : 0));
        }
        catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public void setLightPortParameters(Node node, int port, int pwmMin, int pwmMax, int dimMin, int dimMax) {
        try {
            Bidib.getInstance().getAccessoryNode(node).setConfig(
                new LcConfig(LcOutputType.LIGHTPORT, port, pwmMin, pwmMax, dimMax, dimMin));
        }
        catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public void setServoPortParameters(Node node, int port, int trimDown, int trimUp, int speed) {
        try {
            Bidib.getInstance().getAccessoryNode(node).setConfig(
                new LcConfig(LcOutputType.SERVOPORT, port, trimDown, trimUp, speed, 0));
        }
        catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public void setSpeed(
        Node node, int address, SpeedSteps speedSteps, Integer speed, Direction direction, BitSet activeFunctions,
        BitSet functions) {
        LOGGER.info("set speed, speedSteps: {}, speed: {}, direction: {}, activeFunctions: {}, functions: {}",
            speedSteps, speed, direction, activeFunctions, functions);
        try {
            Bidib.getInstance().getCommandStationNode(node).setDrive(address, speedSteps.getType(), speed,
                direction.getType(), activeFunctions, functions);
        }
        catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public void startAccessory(Node node, Accessory accessory, int aspect) {
        try {
            Bidib.getInstance().getAccessoryNode(node).setAccessoryState(accessory.getId(), aspect);
        }
        catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public void startMacro(Node node, Macro macro) {
        try {
            transferMacro(node, macro);
            Bidib.getInstance().getAccessoryNode(node).handleMacro(macro.getId(), LcMacroOperationCode.START);
        }
        catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public void stopMacro(Node node, Macro macro) {
        try {
            Bidib.getInstance().getAccessoryNode(node).handleMacro(macro.getId(), LcMacroOperationCode.OFF);
        }
        catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    public void transferMacro(Node node, Macro macro) {
        try {
            AccessoryNode accessoryNode = Bidib.getInstance().getAccessoryNode(node);

            if (accessoryNode != null) {
                // delete the macro
                LcMacroState macroState = accessoryNode.handleMacro(macro.getId(), LcMacroOperationCode.DELETE);
                LOGGER.debug("Delete macro returned: {}", macroState);

                byte stepNumber = 0;

                // add the steps
                for (Function function : macro.getFunctions()) {
                    if (function != null) {
                        byte delay = (byte) 255;
                        byte outputNumber = 0;
                        byte value = 0;

                        if (function instanceof DelayFunction) {
                            outputNumber = (byte) ((DelayFunction) function).getDelay();
                        }
                        else if (function instanceof FlagFunction) {
                            Flag flag = ((FlagFunction) function).getFlag();

                            if (flag != null) {
                                outputNumber = (byte) flag.getId();
                            }
                        }
                        else if (function instanceof InputFunction) {
                            Port port = ((InputFunction) function).getInput();

                            if (port != null) {
                                outputNumber = (byte) port.getId();
                            }
                        }
                        else if (function instanceof MacroFunction) {
                            outputNumber = (byte) ((MacroFunction) function).getMacroId();
                        }
                        else if (function instanceof PortAction) {
                            delay = (byte) ((PortAction) function).getDelay();

                            Port port = ((PortAction) function).getPort();

                            if (port != null) {
                                outputNumber = (byte) port.getId();
                            }
                            if (function instanceof ServoPortAction) {
                                value = (byte) ((ServoPortAction) function).getValue();
                            }
                        }
                        else if (function instanceof RandomDelayFunction) {
                            outputNumber = (byte) ((RandomDelayFunction) function).getMaximumValue();
                        }

                        BidibStatus status = function.getAction();

                        accessoryNode.setMacro(new LcMacro((byte) macro.getId(), stepNumber++, delay,
                            getOutputType(function), outputNumber, status != null ? status.getType() : null, value));
                        if (function instanceof MacroFunction
                            && ((MacroFunction) function).getAction() == MacroStatus.END) {
                            break;
                        }
                    }
                }
                // if there's a free entry add the end_of_macro step
                if (macro.getFunctionCount() < macro.getFunctionSize()) {
                    accessoryNode.setMacro(new LcMacro((byte) macro.getId(), stepNumber++, (byte) 255,
                        LcOutputType.END_OF_MACRO, (byte) 0, null, (byte) 0));
                }

                // set the execution time
                accessoryNode
                    .setMacroParameter(macro.getId(), BidibLibrary.BIDIB_MACRO_PARA_START_CLK, MACRO_START_OFF);

                accessoryNode.setMacroParameter(macro.getId(), BidibLibrary.BIDIB_MACRO_PARA_SLOWDOWN, (byte) macro
                    .getSpeed());
                accessoryNode.setMacroParameter(macro.getId(), BidibLibrary.BIDIB_MACRO_PARA_REPEAT, (byte) macro
                    .getCycles());

                for (StartCondition startCondition : macro.getStartConditions()) {
                    if (startCondition instanceof TimeStartCondition) {
                        Calendar time = ((TimeStartCondition) startCondition).getTime();
                        LOGGER.trace("Use time as start condition: {}", time.getTime());
                        if (time != null) {
                            byte[] parameter = new byte[4];
                            int minute = time.get(Calendar.MINUTE);

                            if (((TimeStartCondition) startCondition).getRepeatTime() == MacroRepeatTime.MINUTELY) {
                                minute = 60;
                            }
                            else if (((TimeStartCondition) startCondition).getRepeatTime() == MacroRepeatTime.QUARTER_HOURLY) {
                                minute = 61;
                            }
                            else if (((TimeStartCondition) startCondition).getRepeatTime() == MacroRepeatTime.HALF_HOURLY) {
                                minute = 62;
                            }
                            LOGGER.debug("Value for minute: {}", minute);
                            // the minute
                            parameter[0] = SysClockMessage.getMinute(minute);
                            // the hour
                            MacroRepeatTime repeatTime = ((TimeStartCondition) startCondition).getRepeatTime();
                            parameter[1] =
                                SysClockMessage.getHour(repeatTime != MacroRepeatTime.NONE ? 24 : time
                                    .get(Calendar.HOUR_OF_DAY));
                            // the day of week
                            MacroRepeatDay repeatDay = ((TimeStartCondition) startCondition).getRepeatDay();
                            if (repeatTime == MacroRepeatTime.NONE) {
                                // of no repetition fire on every day
                                repeatDay = MacroRepeatDay.ALL;
                            }
                            parameter[2] = SysClockMessage.getDay(repeatDay.ordinal());
                            // the acceleration factor
                            parameter[3] = SysClockMessage.getAccelerationFactor(0);
                            LOGGER.info("Set macro start clk with parameters: {}", new Object[] { parameter });
                            accessoryNode.setMacroParameter(macro.getId(), BidibLibrary.BIDIB_MACRO_PARA_START_CLK,
                                parameter);
                        }
                    }
                }

            }
        }
        catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    @Override
    public void writeCv(Node node, int cvNumber, int cvValue) {
        boolean vendorEnabled = false;
        try {
            vendorEnabled = Bidib.getInstance().getNode(node).vendorEnable(getUniqueId(node.getUniqueId()));
            if (vendorEnabled) {
                Bidib.getInstance().getNode(node).vendorSet(String.valueOf(cvNumber), String.valueOf(cvValue));
            }
        }
        catch (Exception e) {
            throw new RuntimeException(e);
        }
        finally {
            if (vendorEnabled) {
                try {
                    Bidib.getInstance().getNode(node).vendorDisable();
                }
                catch (Exception e) {
                    throw new RuntimeException(e);
                }
            }
        }
    }

    private void fireStatus(String statusText, int displayDuration) {
        for (CommunicationListener listener : communicationListeners) {
            listener.status(statusText, displayDuration);
        }
    }

    private void fireOpened(String port) {
        for (CommunicationListener listener : communicationListeners) {
            listener.opened(port);
        }
    }

    @Override
    public String toString() {
        StringBuffer sb = new StringBuffer(getClass().getSimpleName());
        sb.append("connected port: ").append(connectedCommPort);
        return sb.toString();
    }
}
