package org.bidib.wizard.mvc.preferences.view.panel;

import java.awt.Dimension;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.util.Collection;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Set;

import javax.swing.BorderFactory;
import javax.swing.ButtonGroup;
import javax.swing.DefaultComboBoxModel;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.JTextField;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;

import org.apache.commons.lang.StringUtils;
import org.bidib.wizard.locale.Resources;
import org.bidib.wizard.mvc.preferences.model.CommPort;
import org.bidib.wizard.mvc.preferences.model.PreferencesModel;
import org.bidib.wizard.mvc.preferences.model.listener.PreferencesModelListener;
import org.bidib.wizard.mvc.preferences.view.listener.PreferencesViewListener;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * This panel displays miscellaneous preferences like selected comm port, path to log files or label files.
 *
 */
public class MiscPanel extends JPanel {
    private static final Logger LOGGER = LoggerFactory.getLogger(MiscPanel.class);

    private static final long serialVersionUID = -5927872456501821169L;

    private final Collection<PreferencesViewListener> listeners = new LinkedList<PreferencesViewListener>();

    private final PreferencesModel model;

    private CommPortRadioPanel commPortRadioPanel;

    public MiscPanel(final PreferencesModel model) {
        this.model = model;
        setBorder(BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), Resources.getString(getClass(),
            "title")
            + ":"));
        setLayout(new GridBagLayout());

        GridBagConstraints c = new GridBagConstraints();

        c.anchor = GridBagConstraints.FIRST_LINE_START;
        c.fill = GridBagConstraints.HORIZONTAL;
        c.gridx = 0;
        c.gridy = 0;
        c.weightx = 0;
        c.weighty = 0;
        c.insets = new Insets(5, 5, 5, 5);
        add(new JLabel(Resources.getString(getClass(), "serialPort") + ":"), c);

        final JComboBox commPorts = new JComboBox();

        commPorts.addActionListener(new ActionListener() {
            @Override
            public void actionPerformed(ActionEvent e) {
                CommPort commPort = (CommPort) commPorts.getSelectedItem();
                LOGGER.debug("The selection of the comm port has changed: {}", commPort);

                fireCommPortChanged(commPort.getName());
            }
        });
        c.gridx += 1;
        c.weightx = 1;
        //        add(commPorts, c);
        commPortRadioPanel = new CommPortRadioPanel(commPorts);
        add(commPortRadioPanel, c);

        // set the comm ports after creation of the radio panel
        setCommPorts(commPorts);
        commPortRadioPanel.setEnabled(!model.isMock());

        c.gridx = 0;
        c.gridy += 1;
        c.weightx = 0;
        add(new JLabel(Resources.getString(getClass(), "labelPath") + ":"), c);

        final JTextField labelPath = new JTextField();

        labelPath.getDocument().addDocumentListener(new DocumentListener() {
            public void changedUpdate(DocumentEvent e) {
                setText();
            }

            public void removeUpdate(DocumentEvent e) {
                setText();
            }

            public void insertUpdate(DocumentEvent e) {
                setText();
            }

            private void setText() {
                model.setLabelPath(labelPath.getText());
            }
        });
        labelPath.setText(model.getLabelPath());
        c.gridx += 1;
        c.weightx = 1;
        add(labelPath, c);

        c.gridx = 0;
        c.gridy += 1;
        c.weightx = 0;
        add(new JLabel(Resources.getString(getClass(), "logFilePath") + ":"), c);

        final JTextField logFilePath = new JTextField();

        logFilePath.getDocument().addDocumentListener(new DocumentListener() {
            public void changedUpdate(DocumentEvent e) {
                setText();
            }

            public void removeUpdate(DocumentEvent e) {
                setText();
            }

            public void insertUpdate(DocumentEvent e) {
                setText();
            }

            private void setText() {
                model.setLogFilePath(logFilePath.getText());
            }
        });
        logFilePath.setText(model.getLogFilePath());
        logFilePath.setPreferredSize(new Dimension(300, logFilePath.getPreferredSize().height));
        c.gridx += 1;
        c.weightx = 1;
        add(logFilePath, c);

        JCheckBox mock = new JCheckBox(Resources.getString(getClass(), "mock"), model.isMock());

        mock.addItemListener(new ItemListener() {
            public void itemStateChanged(ItemEvent e) {
                final boolean mock = e.getStateChange() == ItemEvent.SELECTED;

                commPortRadioPanel.setEnabled(!mock);
                fireMockChanged(mock);

                // set the selected COM port if not mock mode
                if (!mock) {
                    LOGGER.debug("Set the selected comm port.");
                    setCommPorts(commPorts);
                }

            }
        });
        c.gridwidth = 2;
        c.gridx = 0;
        c.gridy += 1;
        c.weightx = 0;
        c.weighty = 1;
        add(mock, c);

        model.addPreferencesModelListener(new PreferencesModelListener() {
            @Override
            public void commPortsChanged() {
                LOGGER.debug("Preferences model listener, commPortsChanged.");
                setCommPorts(commPorts);
            }
        });
    }

    public void addPreferencesViewListener(PreferencesViewListener l) {
        listeners.add(l);
    }

    private void fireCommPortChanged(String commPort) {
        LOGGER.debug("commport has changed: {}", commPort);
        for (PreferencesViewListener l : listeners) {
            l.commPortChanged(commPort);
        }
    }

    private void fireMockChanged(boolean mock) {
        for (PreferencesViewListener l : listeners) {
            l.mockChanged(mock);
        }
    }

    private void setCommPorts(JComboBox commPortsCombo) {
        String activeCommPortName = model.getCommPort();
        LOGGER.debug("The activeCommPortName: {}", activeCommPortName);

        Set<CommPort> commPorts = new HashSet<CommPort>();
        CommPort activeCommPort = null;
        for (CommPort commPort : model.getCommPorts()) {
            commPorts.add(commPort);
            if (activeCommPortName != null && activeCommPortName.equals(commPort.getName())) {
                activeCommPort = commPort;
            }
        }

        LOGGER.debug("Add empty item if someone wants to release the port.");
        CommPort commPortNone = new CommPort(CommPort.NONE);
        commPorts.add(commPortNone);

        commPortsCombo.setModel(new DefaultComboBoxModel(commPorts.toArray(new CommPort[0])));

        if (activeCommPort != null) {
            LOGGER.debug("Select the active commport: {}", activeCommPort);
            commPortRadioPanel.setSelectedPort(activeCommPort);
        }
        else {
            // on linux and mac we must support symlinks ...
            if (StringUtils.isNotBlank(activeCommPortName) && !CommPort.NONE.equals(activeCommPortName)) {
                LOGGER.debug("Set the active symlink: {}", activeCommPortName);
                commPortRadioPanel.setSymlink(activeCommPortName);
            }
            else {
                LOGGER.debug("Select <none> item because no default active port is set.");
                commPortRadioPanel.setSelectedPort(commPortNone);
            }
        }

        // do not call fireCommPortChanged
    }

    private class CommPortRadioPanel extends JPanel implements ActionListener {
        private static final long serialVersionUID = 1L;

        private JComboBox commPorts;

        private JTextField textField;

        private JRadioButton comboButton;

        private JRadioButton textFieldButton;

        public CommPortRadioPanel(JComboBox commPorts) {
            super(new GridBagLayout());
            this.commPorts = commPorts;

            comboButton = new JRadioButton("");
            comboButton.setSelected(true);
            textFieldButton = new JRadioButton("");

            //Group the radio buttons.
            ButtonGroup group = new ButtonGroup();
            group.add(comboButton);
            group.add(textFieldButton);

            textField = new JTextField(20);
            textField.getDocument().addDocumentListener(new DocumentListener() {

                @Override
                public void removeUpdate(DocumentEvent e) {
                    setText();
                }

                @Override
                public void insertUpdate(DocumentEvent e) {
                    setText();
                }

                @Override
                public void changedUpdate(DocumentEvent e) {
                    setText();
                }

                private void setText() {
                    fireCommPortChanged(textField.getText());
                }
            });

            comboButton.addActionListener(CommPortRadioPanel.this);
            textFieldButton.addActionListener(CommPortRadioPanel.this);

            textField.setEnabled(textFieldButton.isSelected());

            GridBagConstraints gbc = new GridBagConstraints();
            gbc.insets = new Insets(0, 0, 5, 0);
            add(comboButton, gbc);

            gbc.gridx = 1;
            gbc.fill = GridBagConstraints.HORIZONTAL;
            gbc.weightx = 1.0;
            add(commPorts, gbc);

            gbc = new GridBagConstraints();
            gbc.gridy = 1;
            add(textFieldButton, gbc);

            gbc.gridx = 1;
            gbc.fill = GridBagConstraints.HORIZONTAL;
            gbc.weightx = 1.0;
            add(textField, gbc);
        }

        private void setSelectedPort(CommPort commPort) {
            comboButton.setSelected(true);
            commPorts.setSelectedItem(commPort);
            actionPerformed(null);
        }

        private void setSymlink(String symlink) {
            textFieldButton.setSelected(true);
            textField.setText(symlink);
            actionPerformed(null);
        }

        @Override
        public void actionPerformed(ActionEvent e) {

            commPorts.setEnabled(comboButton.isSelected() && comboButton.isEnabled());
            textField.setEnabled(textFieldButton.isSelected() && textFieldButton.isEnabled());
        }

        @Override
        public void setEnabled(boolean enabled) {
            super.setEnabled(enabled);

            comboButton.setEnabled(enabled);
            textFieldButton.setEnabled(enabled);
            actionPerformed(null);
        }
    }
}
